'''
Energy_Diversity_Scheme.py, Geoffrey Weal, 28/10/2018

This is one of the Diversity schemes that can be used by the genetic algorithm program. This is the Energy Diversity Scheme.

This scheme works by preventing the population from having clusters with the same energy existing in the population at any one time.

'''

############################################################################################################################################################
###### These Methods are used by other Diversity Scheme. For this reason, these methods have been written to be accessible to those Diversity Schemes ######
############################################################################################################################################################

import os, sys, inspect

###########################################################################################################
########################## Methods Required for the def check_initial_population ##########################
###########################################################################################################

def check_initial_population(self,return_report=False):
	"""
	This definition is responsible for making sure that the initialised population obeys the Energy Diversity Scheme. 

	Here, the clusters in the population are checked to see if they have the same energy. This is needed for after the 
	population has been initally populated with randomly generated clusters and clusters the user has specified.
	
	:param return_report: This indicates if the user want to return a report on which clusters were removed and why (i.e. what clusters it was similar to energetically.) 
	:type  return_report: bool.

	returns:
		* **clusters_to_remove** (*[[int,str],...]*): Contains a list of [the index of cluster in pop, the name of the cluster] that have been removed from the population as they violate the Energy Diversity Scheme.
		* **Energy_Diversity_report** (*{int: [int,...]}*): Contain a more detailed report of what this method has done by returning a dictorary in the format of {name of kept cluster: [list of names of clusters that have been removed because they have the same energy as the cluster that was kept.]}

	"""
	########################################################################################################
	# Sort the indices of clusters in the population by fitness, from highest to lowest.
	cluster_in_pop_index_fitness = sorted(zip(range(len(self.population)),[cluster.fitness for cluster in self.population]), key=lambda x:x[1], reverse=True)
	if (not cluster_in_pop_index_fitness == []) and (not cluster_in_pop_index_fitness[0][1] >= cluster_in_pop_index_fitness[-1][1]):
		print('Error in def check_initial_population in class CNA_Diversity_Scheme, in Energy_Diversity_Scheme.py')
		print('The fitnesses after the sort are the wrong way around')
		print('cluster_in_pop_index_fitness[0][1] (fitness) = '+str(cluster_in_pop_index_fitness[0][1]))
		print('cluster_in_pop_index_fitness[-1][1] (fitness)= '+str(cluster_in_pop_index_fitness[-1][1]))
		print('The fitnesses need to be order from highest to lowest.')
		exit('Check this out. This program will finish without completing.')
	indices_of_pop_sorted_by_fitness = [index for index, fitness in cluster_in_pop_index_fitness]
	########################################################################################################
	# Energy_Diversity_report will give a report of what clusters were removed and why.
	if return_report:
		Energy_Diversity_report = {}
	########################################################################################################
	clusters_to_remove = []
	i1 = 0
	while i1 < len(indices_of_pop_sorted_by_fitness):
		index1 = indices_of_pop_sorted_by_fitness[i1]; cluster1 = self.population[index1]
		i2 = i1 + 1
		while i2 < len(indices_of_pop_sorted_by_fitness):
			index2 = indices_of_pop_sorted_by_fitness[i2]; cluster2 = self.population[index2]
			if abs(round(cluster2.energy - cluster1.energy,12)) < self.minimum_energy_diff:
				name_to_remove  = cluster2.name
				index_to_remove = index2
				# Record information about cluster that were kept and as a result which clusters were removed
				if return_report:
					name_to_keep = cluster1.name; 
					Energy_Diversity_report.setdefault(name_to_keep,[]).append(name_to_remove)
				# add cluster2 to clusters_to_remove and remove it from indices_of_pop_sorted_by_fitness
				clusters_to_remove.append((index_to_remove, name_to_remove))
				del indices_of_pop_sorted_by_fitness[i2]
			else:
				i2 += 1
		i1 += 1
	if return_report:
		Energy_Diversity_report = Energy_Diversity_report.items()
	########################################################################################################
	# Check to see if there are no two clusters in the clusters_to_remove that are the same cluster (same name)
	check_for_duplicate_names = [cluster_name_to_remove for _,cluster_name_to_remove in clusters_to_remove]
	if not len(check_for_duplicate_names) == len(set(check_for_duplicate_names)):
		print('Error in def check_Population in Class Energy_Diversity_Scheme, in Energy_Diversity_Scheme.py')
		print('Two of the clusters name in the removed cluster list are the same')
		print('This indicates that this cluster has been deleted twice?')
		print('List of cluster names in the population: '+str(check_for_duplicate_names))
		print('Check this out.')
		import pdb; pdb.set_trace()
		exit()
	########################################################################################################
	# Sort the indices of the clusters_to_remove from lowest to highest index so we can at some point 
	# remove the clusters with these indices from the self.population.
	clusters_to_remove.sort(key=lambda x:x[0],reverse=False) # Order index from lowest to highest
	########################################################################################################
	# Remove duplicates from the population
	for index_to_remove, name_to_remove in clusters_to_remove[::-1]: # remove index from highest to lowest
		self.population.remove(index_to_remove)
	########################################################################################################
	# Return the results, depending on if the user wants to know which clusters were removed as they were energetic similar to which cluster in same_energies_report
	if return_report:
		return clusters_to_remove, Energy_Diversity_report
	else:
		return clusters_to_remove
	########################################################################################################

########################################################################################################
########################## Methods Required for the def assess_for_violations ##########################
########################################################################################################

class Cluster_Block:
	"""
	This is used by the Remove_Cluster_Due_To_Diversity_Violation definition to store information in an easy way to help the user understand what is going on in this method 

	:param collection: This is the collection to record. This is either the instance of the Population or the Offspring_Pool
	:type  collection: Population or Offspring_Pool
	:param collection_type: This describes if the cluster recorded is in the population (given as 'pop') or the offspring (given as 'off').
	:type  collection_type: str.
	:param index: The position/index of the cluster in the collection.
	:type  index: int

	"""
	def __init__(self,collection,collection_type,index):
		self.collection_type = collection_type
		self.index = index
		self.name = collection[index].name
		self.energy = collection[index].energy
		self.fitness = collection[index].fitness
	def  __repr__(self):
		return str(self.collection_type)+str(self.name)+'('+str(self.index)+', energy = '+str(self.energy)+' eV, fitness = '+str(self.fitness)+')'

def assess_for_violations_message(swap_P_O,remove_O_similar_to_too_many_P,removal_O_P,removal_O_O):
	########################################################################################################
	# Mention offspring that have been removed and due to having a similar energy to which cluster in the population
	# Error with this one
	print('1. Clusters in the Population that have been replaced by Offspring, due to the cluster in the offspring being energetically similar and having a lower energy than a offspring.')
	if len(swap_P_O) > 0:
		print('1. Clusters in the Population that have been replaced by Offspring, due to the cluster in the offspring being energetically similar and having a lower energy than a offspring.')
		for pop_to_swap_out, off_to_swap_in in sorted(swap_P_O,key=lambda x:x[1].name,reverse=False):
			cp_removed_name = pop_to_swap_out.name; cp_removed_energy = pop_to_swap_out.energy; cp_removed_fitness = pop_to_swap_out.fitness
			off_swapped_in_name = off_to_swap_in.name; off_swapped_in_energy = off_to_swap_in.energy; off_swapped_in_fitness = off_to_swap_in.fitness
			print(' --> Population Cluster removed '+str(cp_removed_name)+' (Energy: '+str(round(cp_removed_energy,6))+' eV, fitness = '+str(round(cp_removed_fitness,6))+') {Replaced with Offspring '+str(off_swapped_in_name)+' (Energy: '+str(round(off_swapped_in_energy,6))+' eV, fitness = '+str(round(off_swapped_in_fitness,6))+')}.')
	else:
		print('1. No clusters in the population have been replaced by offspring in the offsprings list')
	print('*****************************')
	########################################################################################################
	# Mention offspring that have been removed and due to having a similar energy to which cluster in the population
	print('2. These offspring have a better energy and are energetically similar to two or more clusters in the population.')
	if len(remove_O_similar_to_too_many_P) > 0:
		print('2. These offspring have a better energy and are energetically similar to two or more clusters in the population.')
		print('   Therefore it is best to remove the offspring to avoid having energetically similar clusters in the population.')
		print('   Below are the offspring that will be removed and the clusters in the population they are energetically similar to.')
		for off_to_swap_out, geo_sim_pops in sorted(remove_O_similar_to_too_many_P,key=lambda x:x[0].name,reverse=False):
			off_removed_name = off_to_swap_out.name; off_removed_energy = off_to_swap_out.energy; off_removed_fitness = off_to_swap_out.fitness
			geo_sim_pop1, geo_sim_pop2 = geo_sim_pops
			geo_sim_pop1_name = geo_sim_pop1.name; geo_sim_pop1_energy = geo_sim_pop1.energy; geo_sim_pop1_fitness = geo_sim_pop1.fitness
			geo_sim_pop2_name = geo_sim_pop2.name; geo_sim_pop2_energy = geo_sim_pop2.energy; geo_sim_pop2_fitness = geo_sim_pop2.fitness
			print(' --> Offspring removed '+str(off_removed_name)+' (Energy: '+str(round(off_removed_energy,6))+' eV; fitness = '+str(round(off_removed_fitness,6))+') {Energetically similar to (at least) clusters in population: '+str(geo_sim_pop1_name)+' (Energy: '+str(round(geo_sim_pop1_energy,6))+' eV, fitness = '+str(round(geo_sim_pop1_fitness,6))+')} and '+str(geo_sim_pop2_name)+' (Energy: '+str(round(geo_sim_pop2_energy,6))+' eV, fitness = '+str(round(geo_sim_pop2_fitness,6))+')}.')
	else:
		print('2. No offspring need to be removed from the offsprings list')
	print('*****************************')
	########################################################################################################
	# Mention offspring that have been removed and due to having a similar energy to which cluster in the population
	print('3. The following offspring were removed from the offspring pool (and because they are energetically similar and have a lower energy to which cluster in the population):')
	if len(removal_O_P) > 0:
		print('3. The following offspring were removed from the offspring pool (and because they are energetically similar and have a lower energy to which cluster in the population):')
		for offspring_removed, cp_due_to in removal_O_P:
			offspring_removed_name = offspring_removed.name; offspring_removed_energy = offspring_removed.energy; offspring_removed_fitness = offspring_removed.fitness
			cp_due_to_name = cp_due_to.name; cp_due_to_energy = cp_due_to.energy; cp_due_to_fitness = cp_due_to.fitness
			print(' --> Offspring '+str(offspring_removed_name)+' (Energy: '+str(offspring_removed_energy)+' eV, fitness = '+str(round(offspring_removed_fitness,6))+') {Similar to Population Cluster '+str(cp_due_to_name)+' (Energy: '+str(round(cp_due_to_energy,6))+' eV, fitness = '+str(round(cp_due_to_fitness,6))+')}.')
	else:
		print('3. No offspring need to be removed from the offsprings list')
	print('*****************************')
	########################################################################################################
	# Mention offspring that have been removed and due to having a similar energy to which other offspring in the Offspring_Pool
	print('4. Offspring that have have been removed for being energetically similar and have a lower energy to the rest of the other Offspring')
	if len(removal_O_O) > 0:
		print('4. Offspring that have have been removed for being energetically similar and have a lower energy to the rest of the other Offspring')
		for offspring_removed, due_to_offspring in removal_O_O:
			offspring_removed_name = offspring_removed.name; offspring_removed_energy = offspring_removed.energy; offspring_removed_fitness = offspring_removed.fitness
			due_to_offspring_name = due_to_offspring.name; due_to_offspring_energy = due_to_offspring.energy; due_to_offspring_fitness = due_to_offspring.fitness
			print(' --> Offspring '+str(offspring_removed_name)+' (Energy: '+str(offspring_removed_energy)+'eV, fitness = '+str(round(offspring_removed_fitness,6))+') {Similar to Offspring '+str(due_to_offspring_name)+' (Energy: '+str(due_to_offspring_energy)+' eV, fitness = '+str(round(due_to_offspring_fitness,6))+')}.')
	else:
		print('4. No offspring have been removed from the rest of the offsprings list for being too similar')
	print('**********************************************************')
	print('**********************************************************')
	########################################################################################################

def assess_for_violations_no_force_replacement(self,offspring_pool):
	"""
	This will assess which offspring to remove before the natural selection process.

	The offspring are assessed against clusters in the population. Offspring are removed from the offspring_pool if:
	* They have an energy equal to another offspring
	* They have an energy equal to another cluster in the population.

	:param offspring_pool: This is the offspring that you want to add eventually using the natural selection process to the population.
	:type  offspring_pool: Offspring_Pool

	"""
	if self.print_details:
		print('**********************************************************')
		print('**********************************************************')
		print('Offspring removed from the Offspring Pool due to violating the Comprehensive Energy Diversity Scheme.')
		print('NOTE: energy difference minimum = '+str(self.minimum_energy_diff)+' eV.')
		print('**********************************************************')
		print('**********************************************************')
	########################################################################################################
	if self.print_details:
		print('Start assessing for Energy Diversity')
	offspring_to_remove = []
	# Check offspring against themselves
	structures_to_consider = [Cluster_Block(offspring_pool,'Offspring',index) for index in range(len(offspring_pool))]
	structures_to_consider.sort(key=lambda cluster: cluster.fitness, reverse=False)
	index1 = 0
	while index1 < len(structures_to_consider):
		offspring1 = structures_to_consider[index1]
		for index2 in range(len(structures_to_consider)-1,index1,-1):
			offspring2 = structures_to_consider[index2]
			energy_diff = abs(round(offspring1.energy - offspring2.energy,12))
			if energy_diff < self.minimum_energy_diff:
				offspring_to_remove.append((offspring2.name,offspring2.index))
				del structures_to_consider[index2]
		index1 += 1
	# check offspring against clusters in the populatino
	for index in range(len(structures_to_consider)):
		offspring = structures_to_consider[index]
		for cluster_in_population in self.population:
			energy_diff = abs(round(offspring.energy - cluster_in_population.energy,12))
			if energy_diff < self.minimum_energy_diff:
				offspring_to_remove.append((offspring.name,offspring.index))
				break
	# Sort by index and continue on
	offspring_to_remove.sort(key=lambda x:x[1],reverse=True)
	force_replacement = []
	return offspring_to_remove, force_replacement

def assess_for_violations_force_replacement(self,offspring_pool):
	"""
	This will assess which offspring to remove before the natural selection process.

	The offspring are assessed against clusters in the population. Offspring are removed from the offspring_pool if:
	* They have an energy equal to another offspring
	* They have an energy equal to another cluster in the population.

	:param offspring_pool: This is the offspring that you want to add eventually using the natural selection process to the population.
	:type  offspring_pool: Offspring_Pool

	"""
	print('**********************************************************')
	print('**********************************************************')
	print('Offspring removed from the Offspring Pool due to violating the Comprehensive Energy Diversity Scheme.')
	print('NOTE: energy difference minimum = '+str(self.minimum_energy_diff)+' eV.')
	print('**********************************************************')
	print('**********************************************************')
	########################################################################################################
	# sort all clusters in population and offspring by fitness
	print('Assessing which structures to remove to comply with the Energy Diversity Scheme')
	structures_to_consider = [Cluster_Block(self.population,'Population',index) for index in range(len(self.population))] 
	structures_to_consider += [Cluster_Block(offspring_pool,'Offspring',index) for index in range(len(offspring_pool))]
	# order fitnesses,From highest to lowest
	structures_to_consider.sort(key=lambda cluster: cluster.fitness, reverse=True)
	if (not structures_to_consider == []) and (not structures_to_consider[0].fitness >= structures_to_consider[-1].fitness):
		print('Error in def assess_for_violations in class CNA_Diversity_Scheme, in Energy_Diversity_Scheme.py')
		print('The fitnesses after the sort are the wrong way around')
		print('structures_to_consider[0].fitness = '+str(structures_to_consider[0].fitness))
		print('structures_to_consider[-1].fitness = '+str(structures_to_consider[-1].fitness))
		print('The fitnesses need to be order from highest to lowest.')
		exit('Check this out. This program will finish without completing.')
	########################################################################################################
	print('Start assessing for Energy Diversity')
	index_higher = 0
	removal_O_P = []; removal_O_O = []
	swap_P_O = []; 
	remove_O_similar_to_too_many_P = []
	# scan though all the clusters in offspring and population in order of fitness
	while index_higher < len(structures_to_consider):
		################################################################################################################
		incountered_population_counter = 0
		cluster_higher = structures_to_consider[index_higher]
		one_instance_of_swap_P_O = []
		removal_O_P_temp = []
		remove_O_similar_to_too_many_P_temp = []
		removal_O_O_temp = []
		# scan though all the other clusters which have higher energies
		indices_to_delete = []
		for index_lower in range(index_higher+1, len(structures_to_consider)):
			cluster_lower = structures_to_consider[index_lower]
			############################################################################################################
			energy_diff = abs(round(cluster_higher.energy - cluster_lower.energy,12))
			if energy_diff < self.minimum_energy_diff:
				# The two clusters being looked at at the moment are geometrically the same. 
				if cluster_higher.collection_type == 'Population':
					# if two clusters in the population are geometrically similar, we have an issue.
					if cluster_lower.collection_type == 'Population':
						print('Error in def assess_for_violations in class CNA_Diversity_Scheme, in Energy_Diversity_Scheme.py')
						print('Two of the clusters in the population are identical by the max  similarity cutt off setting.')
						print('cluster_higher: '+str(cluster_higher))
						print('cluster_lower: '+str(cluster_lower))
						import pdb; pdb.set_trace()
						exit('This program will exit without completing.')
					# If a cluster in the population (and has a higher energy) is geometrically similar to an offspring,
					# then it is better to keep the offspring and remove the offspring
					elif cluster_lower.collection_type == 'Offspring':
						removal_O_P_temp.append([cluster_lower,cluster_higher])
				elif cluster_higher.collection_type == 'Offspring':
					# If an offspring is geometrically similar (and has a higher energy) to a cluster in population,
					# then it is better to replace this cluster in the population with this offspring.
					if cluster_lower.collection_type == 'Population':
						# Take the lowest energy cluster in the population to swap with this offspring.
						if one_instance_of_swap_P_O == []:
							one_instance_of_swap_P_O = [[cluster_lower,cluster_higher]]
						else:
							cluster_lower_old, cluster_higher_old = one_instance_of_swap_P_O[0]
							if not cluster_higher_old.name == cluster_higher.name:
								exit('Somethings up')
							remove_O_similar_to_too_many_P_temp.append([cluster_higher,[cluster_lower_old,cluster_lower]])
							one_instance_of_swap_P_O = []; indices_to_delete = []
							print('Energy Scheme Debugger Marker 1')
							break
					# If an offspring is geometrically similar (and has a higher energy) to an offspring,
					# remove the higher energy offspring
					elif cluster_lower.collection_type == 'Offspring':
						removal_O_O_temp.append([cluster_lower,cluster_higher])
				indices_to_delete.append(index_lower)
		################################################################################################################
		if not remove_O_similar_to_too_many_P_temp == []:
			remove_O_similar_to_too_many_P += remove_O_similar_to_too_many_P_temp
			del structures_to_consider[index_higher]
		else:
			removal_O_P += removal_O_P_temp
			removal_O_O += removal_O_O_temp
			swap_P_O += one_instance_of_swap_P_O
			for index_lower in sorted(indices_to_delete,reverse=True): # remove from highest to lowest index
				del structures_to_consider[index_lower]
			index_higher += 1
			################################################################################################################
	####################################################################################################################
	print('Removing offspring that violate the Energy Diversity Scheme')
	offspring_to_remove = []
	for off_to_remove, pop_to_keep in sorted(removal_O_P+removal_O_O+remove_O_similar_to_too_many_P,key=lambda x:x[0].index,reverse=True): # remove from highest to lowest index
		if offspring_pool[off_to_remove.index].name in [offspring.name for cluster,offspring in swap_P_O]:
			print('Error in def assess_for_violations, in class CNA_Diversity_Scheme, in CNA_Diversity_Scheme.py')
			print('Somehow, this algorithm has identified a structure to be in the offspring_pool and the population')
			print('The structure can only be in either the offspring pool or population.')
			print('Offspring: '+str(offspring_pool[off_to_remove.index].name))
			print('Check this out')
			import pdb; pdb.set_trace()
			exit('This program will end without completing.')
		offspring_to_remove.append((offspring_pool[off_to_remove.index].name,off_to_remove.index))
	####################################################################################################################
	assess_for_violations_message(swap_P_O,remove_O_similar_to_too_many_P,removal_O_P,removal_O_O)
	####################################################################################################################
	# Deal with the offspring that you would like to swap with clusters in the population.
	force_replacement = []
	for pop_to_swap_out, off_to_swap_in in swap_P_O:
		if not off_to_swap_in.name in offspring_pool.get_cluster_names():
			print('Error in assess_for_violations.py in Comprehensive_Energy_Diversity_Scheme_energy.py')
			print('One of the offspring that will be swapped into the population for force_replacement does not exist in the offspring_pool')
			print('Offspring to swap into the population: '+str(swap_P_O))
			print('Offspring in the offspring_pool: '+str(offspring_pool.get_cluster_names()))
			print('Problematic offspring that was incountered: '+str(off_to_swap_in.name))
			print('Check this.')
			import pdb; pdb.set_trace()
			exit('This program will finish without completing.')
		force_replacement.append((pop_to_swap_out.name,off_to_swap_in.name))
	force_replacement.sort(key=lambda x:x[1],reverse=True) # force replace from highest to lowest index
	return offspring_to_remove, force_replacement

################################################################################################################################################################################################################
############################################################### DEBUGGING METHODS ##############################################################################################################################
################################################################################################################################################################################################################

def Check_for_Issue_with_Scheme_with_collection(self,collection):
	"""
	This method allows us to check that the collection does not violate the comprehensive diversity scheme after it has been perform on the collection.

	The collection is either the Population of Offspring_Pool.

	:param collection: This is the collection to record. This is either the instance of the Population or the Offspring_Pool
	:type  collection: Population or Offspring_Pool
		
	"""
	energies = [cluster.energy for cluster in collection]
	similar_energy_clusters = []
	for index1 in range(len(self.population)):
		for index2 in range(index1+1,len(self.population)):
			energy_diff = round(self.population[index1].energy - self.population[index2].energy,12)
			if abs(energy_diff) < self.minimum_energy_diff:
				similar_energy_clusters.append([index1,index2,abs(energy_diff)])
	if len(similar_energy_clusters) > 0:
		print('*******************************************')
		print('Error with Energy Diversity Scheme.')
		print('There are duplicates in the energy, which should not be happening in the Energy Diversity Scheme.')
		print('*******************************************')
		print('Here is a printout of the clusters in '+str(collection.getname()))
		for cluster in collection:
			print('Cluster '+str(cluster)+'; Energy = '+str(cluster.energy)+' eV.')
		print('*******************************************')
		print('The clusters that have an energy less than the minimum energy diff ('+str(self.minimum_energy_diff)+' eV) are:')
		for index1, index2, abs_energy_diff in similar_energy_clusters:
			cluster1 = self.population[index1]; cluster2 = self.population[index2]
			print('Cluster '+str(cluster1.name)+' (Energy = '+str(cluster1.energy)+' eV); Cluster '+str(cluster2.name)+' (Energy = '+str(cluster2.energy)+' eV); Energy Difference: '+str(abs_energy_diff))
		print('*******************************************')
		print('This likely means there is an error in the Energy Diversity Scheme code. Check this.')
		import pdb; pdb.set_trace()
		exit()

################################################################################################################################################################################################################
################################################################################################################################################################################################################
################################################################################################################################################################################################################






import multiprocessing
# We must import this explicitly, it is not imported by the top-level
# multiprocessing module.
import multiprocessing.pool
import time

from random import randint


class NoDaemonProcess(multiprocessing.Process):
    # make 'daemon' attribute always return False
    def _get_daemon(self):
        return False
    def _set_daemon(self, value):
        pass
    daemon = property(_get_daemon, _set_daemon)

# We sub-class multiprocessing.pool.Pool instead of multiprocessing.Pool
# because the latter is only a wrapper function, not a proper class.
class MyPool(multiprocessing.pool.Pool):
    Process = NoDaemonProcess


#!/usr/bin/python
'''
Geoffrey Weal, remove_lock_files.py, 10/02/2021

This program is designed to remove all the ga_running.lock files from subdirectories that this program is run from.
'''
import os, sys

lock_name = 'ga_running.lock'

path = os.getcwd()
for dirpath, dirs, files in os.walk(path):
	dirs.sort()
	if ('MDRun.py' in files and dirpath.split('/')[-1].startswith('Trial')) or ('run_eon.py' in files):
		if lock_name in files:
			print('Removing '+lock_name+' file in '+str(dirpath))
			os.remove(dirpath+'/'+lock_name)
		dirs[:] = []
		files[:] = []
	if ('Run.py' in files and dirpath.split('/')[-1].startswith('Trial')):
		if lock_name in files:
			print('Removing '+lock_name+' file in '+str(dirpath))
			os.remove(dirpath+'/'+lock_name)
		dirs[:] = []
		files[:] = []


import atexit
from Organisms.GA.Lock import Lock_Remove

def exit_handler():
	print('Unlocking the genetic algorithm program before exitting program.')
	Lock_Remove()

def add_to_exitting_procedure():
	atexit.register(exit_handler)


import os, shutil
from Organisms.GA.Collection import Collection

class Offspring_Pool(Collection):
	"""
	This class is designed to hold the offspring that are made during the Organisms program.

	:param name: The name of the Offspring_Pool. The names of all the collections should be different to prevent confusion, but this shouldn't affect how this program works.
	:type  name: str.
	:param offspring_pool_size: The maximum number of clusters in the collection
	:type  offspring_pool_size: int 

	"""
	def __init__(self,name,offspring_pool_size):
		Collection.__init__(self,name,offspring_pool_size,have_database=False)

	def sort_by_fitness(self):
		"""
		Sort the offspring in the Offspring_Pool by fitness
		"""
		self.clusters.sort(key=lambda cluster: cluster.fitness, reverse=True)

	def clean(self):
		"""
		Remove all the clusters in the collection.
		"""
		cleaned_offspring_names = []
		for index in range(len(self)-1,-1,-1):
			offspring_name = self[index].name
			cleaned_offspring_names.append(offspring_name)
			del self[index]
		return cleaned_offspring_names



#!/usr/bin/python
'''
Geoffrey Weal, make_finish_files.py, 10/02/2021

This program is designed to place finish files in all subdirectories that also include a Run.py file
'''
import os, sys

def make_finish_file(path):
	print('Putting finish file in '+str(path))
	finish_file = open(path+"/"+"finish", "w")
	finish_file.write('')
	finish_file.close()

path = os.getcwd()

for dirpath, dirs, files in os.walk(path):
	dirs.sort()
	if 'Run.py' in files:
		if any([(dir_name.startswith('Trial') and dir_name.replace('Trial','').isdigit()) for dir_name in dirs]):
			continue
		make_finish_file(dirpath)
		dirs[:] = []
		files[:] = []

	if 'MDRun.py' in files:
		trials = [dirname for dirname in dirs if dirname.replace('Trial','').isdigit()]
		for trial in trials:
			make_finish_file(dirpath+'/'+trial)
		dirs[:] = []
		files[:] = []

	if 'run_eon.py' in files:
		make_finish_file(dirpath)
		dirs[:] = []
		files[:] = []


from ase import Atoms
from ase.io import read as ase_read
from ase.build import add_adsorbate
from ase.data import covalent_radii, vdw_radii, atomic_numbers
from ase.constraints import FixAtoms, FixedLine, FixInternals, ExternalForce
import numpy as np
from math import pi
from ase.calculators.emt import EMT
from ase.optimize import FIRE

class MyConstraint:
    """Constrain an atom to move along a given direction only."""
    def __init__(self, a, direction):
        self.a = a
        self.dir = direction / sqrt(np.dot(direction, direction))

    def adjust_positions(self, atoms, newpositions):
        step = newpositions[self.a] - atoms.positions[self.a]
        step = np.dot(step, self.dir)
        newpositions[self.a] = atoms.positions[self.a] + step * self.dir

    def adjust_forces(self, atoms, forces):
        forces[self.a] = self.dir * np.dot(forces[self.a], self.dir) + 10.0*direction

def get_xy_center(atoms):
	x_min = float(min(atoms.positions,key=lambda position: position[0])[0])
	x_max = float(max(atoms.positions,key=lambda position: position[0])[0])
	y_min = float(min(atoms.positions,key=lambda position: position[1])[1])
	y_max = float(max(atoms.positions,key=lambda position: position[1])[1])
	x_center = (x_max - x_min)/2.0
	y_center = (y_max - y_min)/2.0
	return np.array([x_center, y_center])

def get_distance(atom1,atom2):
	diff_x = atom1.x - atom2.x
	diff_y = atom1.y - atom2.y
	diff_z = atom1.z - atom2.z
	distance = (diff_x**2.0+diff_y**2.0+diff_z**2.0)**0.5
	return distance

class Surface:

	def __init__(self,surface_details):
		# Details about the surface
		if surface_details in [None,'none','None', {}]:
			self.surface = None
			self.place_cluster_where = None
		elif isinstance(surface_details['surface'], Atoms):
			self.surface = surface_details['surface']
		elif isinstance(surface_details['surface'], str):
			self.surface = ase_read(surface_details['surface'])
		else:
			exit('Error')
		if not self.surface == None:
			self.surface.center()
			self.place_cluster_where = surface_details['place_cluster_where']

	def adjust_z_axis(self, cluster):
		if self.surface == None:
			return
		'''
		surface_constraints = []
		for atom in self.surface:
			surface_constraints.append(MyConstraint(atom.index,(0,0,1)))
		'''
		surface_constraints = []
		#surface_constraints.append(ExternalForce(0, 0, 10))
		all_bonds = []
		all_angles = []
		all_dihedrals = []
		for index_1 in range(len(self.surface)):
			for index_2 in range(index_1+1,len(self.surface)):
				bonds_distance = get_distance(self.surface[index_1],self.surface[index_2])
				bond = [bonds_distance, [index_1,index_2]]
				all_bonds.append(bond)
				"""
				for index_3 in range(index_2+1,len(self.surface)): 
					angle_indices = [[index_1,index_2,index_3],[index_3,index_1,index_2],[index_2,index_3,index_1]]
					for angle_indice in angle_indices:
						angle = [self.surface.get_angle(*angle_indice) * pi / 180, angle_indice]
						all_angles.append(angle)
					'''
					for index_4 in range(index_3+1,len(self.surface)):  
						dihedral_indices = [[index_1,index_2,index_3,index_4],[index_1,index_2,index_3,index_4],[index_1,index_2,index_3,index_4]]
						for dihedral_indice in dihedral_indices:
							dihedral = [self.surface.get_dihedral(*dihedral_indice) * pi / 180, dihedral_indice]
							all_dihedrals.append(dihedral)
					'''
				"""
		surface_constraints.append(FixInternals(bonds=all_bonds, angles=all_angles, dihedrals=all_dihedrals))
		#self.surface.set_constraint(surface_constraints)

		c = FixAtoms(indices=range(len(self.surface),len(self.surface)+len(cluster)))
		cluster_constraints = [c]
		#cluster.set_constraint(cluster_constraints)

		for atom in self.surface:
			atom.z -= 40.0

		system = self.surface + cluster
		system.set_calculator(EMT())
		system.set_constraint(surface_constraints+cluster_constraints)
		dyn = FIRE(system)
		converged = False
		import pdb; pdb.set_trace()
		try:
			dyn.run(fmax=0.01,steps=5000)
			converged = dyn.converged()
			if not converged:
				import os
				name = os.path.basename(os.getcwd())
				errorMessage = 'The optimisation of cluster ' + name + ' did not optimise completely.'
				print(errorMessage, file=sys.stderr)
				print(errorMessage)
		except:
			print('Local Optimiser Failed for some reason.')
		import pdb; pdb.set_trace()



	def get_surface(self, cluster):
		if self.surface == None:
			return
		if self == None:
			return
		# find center of xy plane in 
		if self.place_cluster_where == None:
			cluster_xy_center = get_xy_center(cluster)
			surface_xy_center = get_xy_center(self.surface)
			xy_distance_diff = cluster_xy_center - surface_xy_center
			xy_distance_diff = np.append(xy_distance_diff,0.0)
			for atom in self.surface:
				atom.position += xy_distance_diff
		else:
			for index in range(len(self.surface)):
				self.surface[index].x = float(self.surface[index].x) + self.place_cluster_where[0]
				self.surface[index].y = float(self.surface[index].y) + self.place_cluster_where[1]

		self.adjust_z_axis(cluster)
			


#!/usr/bin/python
'''
Geoffrey Weal, Run_mass_submitSL_slurm.py, 10/02/2021

This program is designed to submit all sl files called mass_submit.sl to slurm

'''
print('###########################################################################')
print('###########################################################################')
print('Run_mass_submitSl_slurm.py')
print('###########################################################################')
print('This program is designed to submit all your mass_submit.sl scripts appropriately to slurm.')
print('###########################################################################')
print('###########################################################################')

import os, time, sys
import subprocess

# ------------------------------------------------------
# These variables can be changed by the user.
Max_jobs_in_queue_at_any_one_time = 10000
time_to_wait_before_next_submission = 20.0
time_to_wait_max_queue = 60.0

time_to_wait_before_next_submission_due_to_temp_submission_issue = 10.0
number_of_consecutive_error_before_exitting = 20
# ------------------------------------------------------

if len(sys.argv) > 1:
    wait_between_submissions = str(sys.argv[1]).lower()
    if wait_between_submissions in ['t','true']:
        wait_between_submissions = True
    elif wait_between_submissions in ['f','false']:
        wait_between_submissions = False
    else:
        print('If you pass this program an argument, it must be either: ')
        print('    t, true, True: will wait 1 minute between submitting jobs')
        print('    f, false. False: will not wait between submitting jobs')
        print('If no argument is entered, the default is given as True')
        exit('This program will exit without running')
else:
    wait_between_submissions = True

if wait_between_submissions == True:
    print('This program will wait one minute between submitting jobs.')
else:
    print('This program will not wait between submitting jobs.')
path = os.getcwd()

def countdown(t):
    print('Will wait for ' + str(float(t)/60.0) + ' minutes, then will resume Pan submissions.\n')
    while t:
        mins, secs = divmod(t, 60)
        timeformat = str(mins) + ':' + str(secs)
        #timeformat = '{:02d}:{:02d}'.format(mins, secs)
        sys.stdout.write("\r                                                                                   ")
        sys.stdout.flush()
        sys.stdout.write("\rCountdown: " + str(timeformat))
        sys.stdout.flush()
        time.sleep(1)
        t -= 1
    print('Resuming Pan Submissions.\n')

def myrun(cmd):
    proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout_lines = [line for line in iter(proc.stdout.readline,'')]
    #import pdb; pdb.set_trace()
    #proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    #stdout_lines = [line for line in io.TextIOWrapper(proc.stdout, encoding="utf-8")]
    """from http://blog.kagesenshi.org/2008/02/teeing-python-subprocesspopen-output.html"""
    '''
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout = []
    while True:
        line = p.stdout.readline()
        stdout.append(line)
        if line == '' and p.poll() != None:
            break
    '''
    return ''.join(stdout_lines)

###########################################################################################
###########################################################################################
###########################################################################################

def get_number_to_trials_that_will_be_submitted_by_mass_submitSL(dirpath):
    path_to_mass_submitSL = dirpath+'/mass_submit.sl'
    with open(path_to_mass_submitSL,'r') as mass_submitSL:
        for line in mass_submitSL:
            if '#SBATCH --array=' in line:
                trials = line.replace('#SBATCH --array=','')
                no_of_trials_that_will_be_submitted = 0
                worked_successfully = True
                trial_limits_commands = trials.rstrip().split(',')
                for trial_limits in trial_limits_commands:
                    if trial_limits.count('-') == 1:
                        trial_limits = trial_limits.split('-')
                        if len(trial_limits) == 2 and trial_limits[0].isdigit() and trial_limits[1].isdigit():
                            no_of_trials_that_will_be_submitted =+ int(trial_limits[1]) - int(trial_limits[0]) + 1
                        else:
                            worked_successfully = False
                            break
                    elif trial_limits.isdigit():
                        no_of_trials_that_will_be_submitted += 1
                    else:
                        worked_successfully = False
                        break
                if worked_successfully:
                    return no_of_trials_that_will_be_submitted
                else:
                    print('========================================================')
                    print('Error in submitting: '+str(path_to_mass_submitSL))
                    print('One of the clusters in the array to be submitted is not a integer or is entered incorrectly.')
                    print()
                    print(line)
                    print()
                    print('Check this line in your submit.sl script')
                    print('This program will now exit')
                    exit ('========================================================')         
    print('Error in def get_number_to_trials_that_will_be_submitted_by_mass_submitSL, in Run_mass_submitSl_slurm.py script, found in the folder SubsidiaryPrograms in this GA program.')
    print('The mass_submit.sl script found in '+str(dirpath)+' does not have the line that starts with "#SBATCH --array=" in the script.')
    print('Just check this script to make sure everything is all good.')
    import pdb; pdb.set_trace()
    print('This program will finish')
    exit()

command = "squeue -r -u $USER"
def check_max_jobs_in_queue_after_next_submission(dirpath):
    while True:
        text = myrun(command)
        nlines = len(text.splitlines())-1
        if not (nlines == -1):
            break
        else:
            print('Could not get the number of jobs in the slurm queue. Retrying to get this value.')
    number_of_trials_to_be_submitted = get_number_to_trials_that_will_be_submitted_by_mass_submitSL(dirpath)
    if nlines > Max_jobs_in_queue_at_any_one_time - number_of_trials_to_be_submitted:
        return True, nlines
    else:
        return False, nlines

###########################################################################################
###########################################################################################
###########################################################################################

# Check to make sure the array line in all mass_submit.sl scripts is there and that none of the mass_submission script submit more than Max_jobs_in_queue_at_any_one_time
# into the queue. 
print('-----------------------------------------------')
print('Checking to make sure that the array line in all mass_submit.sl scripts is there and that none of the mass_submission script submit more than Max_jobs_in_queue_at_any_one_time into the queue.')
for (dirpath, dirnames, filenames) in os.walk(path):
    dirnames.sort()
    if 'mass_submit.sl' in filenames:
        no_of_trials_that_will_be_submitted = get_number_to_trials_that_will_be_submitted_by_mass_submitSL(dirpath)
        if no_of_trials_that_will_be_submitted > Max_jobs_in_queue_at_any_one_time:
            print('Issue: The number of Trials that you want to submit is greater the number of trials you are allowed to submit at any one time.')
            print('Number of jobs that will be submitted by '+str(dirpath)+'/mass_submit.sl: '+str(no_of_trials_that_will_be_submitted))
            print('Maximum number of trials that can be submitted into slurm: '+str(Max_jobs_in_queue_at_any_one_time))
            print('Consider doing either:')
            print('    1) Using the scripts Create_submitSL_slurm.py and Run_submitSL_slurm.py to run your jobs. This will submit your jobs individually and has better control over submitting this large number of trials within slurm without causing issues with slurm.')
            print('    2) Contact your slurm technical support about increasing the maximum number of jobs that can be found in the queue at any one time.')
            print('This program will exit without doing anything.')
            exit()
        dirnames[:] = []
        filenames[:] = []
print('All clear to submit your mass_submit.sl scripts.')
print('-----------------------------------------------')
print('*****************************************************************************')
print('*****************************************************************************')
print('Submitting mass_submit.sl scripts to slurm.')

# Time to submit all the GA scripts! Lets get this stuff going!
submitting_command = "sbatch mass_submit.sl"
for (dirpath, dirnames, filenames) in os.walk(path):
    dirnames.sort()
    if 'mass_submit.sl' in filenames:
        # determine if it is the right time to submit jobs
        print('*****************************************************************************')
        while True:
            reached_max_jobs, number_in_queue = check_max_jobs_in_queue_after_next_submission(dirpath)
            if reached_max_jobs:
                print('-----------------------------------------------------------------------------')
                print('You can not have any more jobs in the queue before submitting the mass_sub. Will wait a bit of time for some of them to complete')
                print('Number of Jobs in the queue = '+str(number_in_queue))
                countdown(time_to_wait_before_next_submission)
                print('-----------------------------------------------------------------------------')
            else:
                print('The number of jobs in the queue currently is: '+str(number_in_queue))
                break
        # submit the jobs
        os.chdir(dirpath)
        name = dirpath.replace(path, '').split('/', -1)[1:]
        name = "_".join(str(x) for x in name)
        print("Submitting " + str(name) + " to slurm.")
        error_counter = 0
        while True:
            if error_counter == number_of_consecutive_error_before_exitting:
                break
            else:
                proc = subprocess.Popen(submitting_command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                if not proc.wait() == 0:
                    error_counter += 1
                    if error_counter == number_of_consecutive_error_before_exitting:
                        print('----------------------------------------------')
                        print('Error in submitting submit script to slurm.')
                        print('I got '+str(number_of_consecutive_error_before_exitting)+" consecutive errors. Something must not be working right somewhere. I'm going to stop here just in case something is not working.")
                        print('')
                        print('The following mass_submit.sl scripts WERE NOT SUBMITTED TO SLURM')
                        print('')
                    else:
                        stdout, stderr = proc.communicate()
                        print('----------------------------------------------')
                        print('Error in submitting submit script to slurm. This error was:')
                        print(stderr)
                        print('Number of consecutive errors: '+str(error_counter))
                        print('Run_mass_submitSL_slurm.py will retry submitting this job to slurm after '+str(time_to_wait_before_next_submission_due_to_temp_submission_issue)+' seconds of wait time')
                        print('----------------------------------------------')
                        countdown(time_to_wait_before_next_submission_due_to_temp_submission_issue)
                else:
                    break
        if error_counter == number_of_consecutive_error_before_exitting:
            print(dirpath)
        elif not wait_between_submissions:
            pass # do not wait any time before proceeding to the next submission.
        else:
            reached_max_jobs, number_in_queue = check_max_jobs_in_queue_after_next_submission(dirpath)
            print('The number of jobs in the queue after submitting job is currently is: '+str(number_in_queue))
            print('Will wait for '+str(time_to_wait_max_queue)+' to give time between consecutive submissions')
            countdown(time_to_wait_max_queue)
            print('*****************************************************************************')
        dirnames[:] = []
        filenames[:] = []

if error_counter == number_of_consecutive_error_before_exitting:
    print('----------------------------------------------')
    print()
    print('"Run_mass_submitSL_slurm.py" will finish WITHOUT HAVING SUBMITTED ALL JOBS.')
    print()
    print('*****************************************************************************')
    print('NOT ALL mass_submit.sl SCRIPTS WERE SUBMITTED SUCCESSFULLY.')
    print('*****************************************************************************') 
else:
    print('*****************************************************************************')
    print('*****************************************************************************')
    print('*****************************************************************************')
    print('All mass_submit.sl scripts have been submitted to slurm successfully.')
    print('*****************************************************************************')
    print('*****************************************************************************')
    print('*****************************************************************************')


from Organisms.GA.Fitness_Operators.Fitness_Operator import Fitness_Operator
from Organisms.GA.Fitness_Operators.Fitness_Function import Fitness_Function
from Organisms.GA.Fitness_Operators.Energetic_Fitness_Contribution import get_lowest_and_highest_energies_from_collections, get_energetic_fitness_contribution

class Energy_Fitness_Operator(Fitness_Operator):
	"""
	The Fitness class is designed to determine and assign the appropriate fitness value to clusters created during the genetic algorithm. 
	The fitnesses are based on the energies of the clusters in the offspring and the population. 

	There are only two methods that need to be written for the predation operator that is to be used. These are:
		* assign_population_fitnesses
		* assign_all_fitnesses

	:param fitness_information: This is all the information that is needed about the fitness class
	:type fitness_information: dict.
	:param predation_operator: This is the predation operator that this fitness class will take information from if needed to obtain a fitness.
	:type  predation_operator: Organisms.GA.Predation_Operator
	:param population: The population to assign fitnesses to
	:type population: Organisms.GA.Population
	:param print_details: Print the details of the energy fitness operator. True if yes, False if no
	:type  print_details: bool

	"""
	def __init__(self, fitness_information, predation_operator, population, print_details):
		super().__init__(fitness_information, population, print_details)
		self.energy_fitness_options(fitness_information, predation_operator)

	def energy_fitness_options(self, fitness_information, predation_operator):
		"""
		This method is designed to set the variables requires for this fitness.

		:param fitness_information: This is all the information that is needed about the fitness class
		:type fitness_information: dict.
		:param predation_operator: This is the predation operator that this fitness class will take information from if needed to obtain a fitness.
		:type  predation_operator: Organisms.GA.Predation_Operator
		
		"""
		#self.predation_switch = predation_operator.Predation_Switch

		self.fitness_switch = fitness_information['Fitness Operator']
		#self.use_predation_information = fitness_information['Use Predation Information']

		self.previous_max_energy = -float('inf')
		self.previous_min_energy = float('inf')
		self.energy_fitness_function = Fitness_Function(**fitness_information['fitness_function'])

	####################################################################################
	################## Method for assign_initial_population_fitnesses ##################
	####################################################################################

	def assign_initial_population_fitnesses(self):
		"""
		This method is designed to assign fitness values to the clusters of the population only at the start of the GA, when the population has been initialised

		"""
		self.__assign_population_fitnesses(0)

	####################################################################################
	################## Method for assign_resumed_population_fitnesses ##################
	####################################################################################

	def assign_resumed_population_fitnesses(self,resume_from_generation): 
		"""
		This method is designed to assign fitness values to the clusters of the population only at the beginning of a resumed GA. 

		:param resume_from_generation: The current generation to resume from.
		:type  resume_from_generation: int
		
		"""
		self.__assign_population_fitnesses(resume_from_generation)

	############################################################################################################
	################## Method for assign_all_fitnesses_before_assess_against_predation_operator ##################
	############################################################################################################

	def assign_all_fitnesses_before_assess_against_predation_operator(self,all_offspring_pools,current_generation_no):
		"""
		This method is to be used in the GA program. 
		This will assign all the fitnesses of all clusters in the current generation (population and offspring) before the offspring are assessed to understand if they violate the predation operator (i.e. an offspring is Class 1 similar to a cluster in the population, or another offspring).

		See the description given for the "assign_all_fitnesses" def on what the crux of this method is.

		If you write your own diversiy and fitness classes, you do not need to implement this method in your fitness class.

		:param all_offspring_pools: list of all the offspring pools of offspring to assign fitness values to.
		:type  all_offspring_pools: list of Organisms.GA.Offspring_Pool
		:param current_generation_no: The current generation
		:type  current_generation_no: int

		"""
		# At this point:
		#	- self.previous_max_energy = -float('inf'), and 
		#	- self.previous_min_energy = float('inf')
		# Meaning that the fitnesses of all clusters in the population, and all the offspring in the all_offspring_pools, will be calculated and assigned to all clusters.
		self.__assign_all_fitnesses(all_offspring_pools)

	###########################################################################################################
	################## Method for assign_all_fitnesses_after_assess_against_predation_operator ##################
	###########################################################################################################

	def assign_all_fitnesses_after_assess_against_predation_operator(self,all_offspring_pools,current_generation_no, offspring_to_remove):
		"""
		This method is to be used in the GA program. 
		This will assign all the fitnesses of all clusters in the current generation (population and offspring) after the offspring are assessed to understand if they violate the predation operator (i.e. an offspring is Class 1 similar to a cluster in the population, or another offspring).

		See the description given for the "assign_all_fitnesses" def on what the crux of this method is.

		If you write your own diversiy and fitness classes, you do not need to implement this method in your fitness class.

		:param all_offspring_pools: All of the offspring_pools
		:type  all_offspring_pools: list of Organisms.GA.Offspring_Pool
		:param current_generation_no: The current generation
		:type  current_generation_no: int
		:param offspring_to_remove: a list of all the names of clusters in the offspring that will be removed.
		:type  offspring_to_remove: list of int

		"""
		# At this point:
		#	- self.previous_max_energy may not be equal to -float('inf'), and 
		#	- self.previous_min_energy may not be equal to float('inf')
		# Meaning that the fitnesses of all clusters in the population, and all the offspring in the all_offspring_pools, may be recalculated if these energy value have changed by removing clusters due to violating the predation operator
		self.__assign_all_fitnesses(all_offspring_pools)

	#############################################################################################
	################## Method for assign_all_fitnesses_after_natural_selection ##################
	#############################################################################################

	def assign_all_fitnesses_after_natural_selection(self,current_generation_no):
		"""
		This method is to be used in the GA program. 
		This will assign all the fitnesses to all the clusters in the population before the offspring are assessed to understand if they violate the predation operator (i.e. an offspring is Class 1 similar to a cluster in the population, or another offspring).

		See the description given for the "assign_population_fitnesses" def on what the crux of this method is.

		If you write your own diversiy and fitness classes, you do not need to implement this method in your fitness class.

		:param current_generation_no: The current generation
		:type  current_generation_no: int

		"""
		# At this point:
		#	- self.previous_max_energy may not be equal to -float('inf'), and 
		#	- self.previous_min_energy may not be equal to float('inf')
		# Meaning that the fitnesses of all clusters in the population, and all the offspring in the all_offspring_pools, may be recalculated if these energy value have changed by removing clusters due to natural selection.
		self.__assign_population_fitnesses(current_generation_no)
		self.previous_max_energy = -float('inf')
		self.previous_min_energy = float('inf')


	#####################################################################################
	################## Subsidiary Methods for the Energy Fitness Class ##################
	#####################################################################################

	def __assign_population_fitnesses(self,current_generation): 
		"""
		This method is designed to assign fitness values to the clusters of the population only. 

		:param current_generation_no: The current generation.
		:type  current_generation_no: int
		
		"""
		min_energy, max_energy = get_lowest_and_highest_energies_from_collections(self.population,None)
		if not ((min_energy == self.previous_min_energy) and (max_energy == self.previous_max_energy)):
			self.__assign_fitnesses_to_clusters_in_collection(self.population, min_energy, max_energy)

	def __assign_all_fitnesses(self,all_offspring_pools): 
		"""
		This method is designed to assign fitness values of the clusters that need to be assigned or reassigned a new fitness value to what it 
		should be between the processes of creating the offspring and the natrual selection processes. 

		:param all_offspring_pools: This has been designed in advance to be able to process all the different offspring pool if ever the genetic algorithm is designed with multiple offspring_pools. Therefore, this is a list of all the offspring_pools created during a single generation.
		:type  all_offspring_pools: [list of offspring_pool]
		:param current_generation_no: The current generation.
		:type  current_generation_no: int

		"""
		if not isinstance(all_offspring_pools,list):
			all_offspring_pools = [all_offspring_pools]
		min_energy, max_energy = get_lowest_and_highest_energies_from_collections(self.population, all_offspring_pools)
		if not ((min_energy == self.previous_min_energy) and (max_energy == self.previous_max_energy)):
			self.__assign_fitnesses_to_clusters_in_collection(self.population, min_energy, max_energy)
			for offspring_pool in all_offspring_pools:
				self.__assign_fitnesses_to_clusters_in_collection(offspring_pool, min_energy, max_energy)
			self.previous_max_energy = max_energy
			self.previous_min_energy = min_energy

			
	def __assign_fitnesses_to_clusters_in_collection(self, collection, min_energy, max_energy):
		"""
		This definition will attach the fitness value associated to each cluster in the collection object provided by the user.

		See `Theoretical study of Cu-Au nanoalloy clusters using a genetic algorithm, Darby et al., 116, 1536 (2002); doi: 10.1063/1.1429658 <http://dx.doi.org/10.1063/1.1429658>`_, page 1538 about this fitness equation, including the definition of max_energy and min_energy.

		:param collection: This is the collection object that we would like to assign.
		:type  collection: Organisms.GA.Collection
		:param min_energy: This is the lowest energy of the current lowest energetic structure across all the collections in the GA (e.g. [population, offspring]).
		:type  min_energy: float
		:param max_energy: This is the highest energy of the current highest energetic structure across all the collections in the GA (e.g. [population, offspring]).
		:type  max_energy: float

		At the end of this method, all the cluster in the collection will have a fitness value attached to them.
		You can see the fitness value that is attahced to a cluster by looking for th "fitness" variable attached to the cluster.

		"""
		for cluster in collection:
			cluster_fitness = get_energetic_fitness_contribution(cluster, max_energy, min_energy, self.energy_fitness_function)
			cluster.fitness = cluster_fitness		

	def add_to_database(self, collection):
		"""
		This method is required by Organisms.GA.Fitness_Operators.Fitness_Operator, but does not do anything

		:param collection: update the fitnesses of clusters in the collection.
		:type  collection: Organisms.GA.Collection.Collection

		"""
		pass

	def remove_from_database(self, cluster_names_to_remove):
		"""
		This method is required by Organisms.GA.Fitness_Operators.Fitness_Operator, but does not do anything.

		:param cluster_names_to_remove: A list of the names of all the clusters to remove from the CNA database
		:type  cluster_names_to_remove: list of ints
		
		"""
		pass

	def reset(self):
		"""
		This method is required by Organisms.GA.Fitness_Operators.Fitness_Operator, but does not do anything
		"""
		return



import os 
from numpy import zeros
from shutil import copyfile, rmtree
from ase.io import read as ase_read
from ase.io import write as ase_write
from ase.db import connect
from ase.visualize import view

from statistics import mean
from multiprocessing import Process
import multiprocessing as mp

from Organisms.GA.Cluster import Cluster
from Organisms.GA.EnergyProfile import tail, remove_end_lines_from_text

class Collection:
	"""
	This is the foundation of the object used to store clusters in the population, the offspring, and for recording clusters made during the genetic algorithm using the GA_Recording_System.py

	:param path: The path that clusters will be written to disk. Default: None, meaning will write to same path as your execution Run.py file. 
	:type  path: str. 
	:param name: Name of the Collection. This can be any name you like, it is just to note what the collection is. All collections should have a unique name if possible to prevent confusion, however this will not break the program.
	:type  name: str. 
	:param size: This is the number of clusters that should be in the collection. In this version of the Organisms program, the number of clusters in the population and made during Creation of Offspring should be consistant throughout the genetic algorithm process.
	:type  size: int 
	:param write_collection_history: This will tell the collection to record a txt file of what clusters were in the collection over the generations Organisms performs. Default: False
	:type  write_collection_history: bool.
	:param write_files_as: This tells the collection if and how to write clusters to the disk (Default: "database"). There are three options:
							
							* "database" if you want the Collection to make a database
							* "xyz" if you want the Collection to make xyz files
							* None, "None" or "none" if you do not want the Collection to make any cluster files
	:type  write_files_as: str.

	"""
	def __init__(self,name,size,path=None,have_database=False,write_collection_history=False,write_cluster_in_RAM=True):
		# Write main details for Collection object.
		self.name = name
		self.size = size
		self.write_cluster_in_RAM = write_cluster_in_RAM
		# sort out if writing clusters to database:
		if self.write_cluster_in_RAM:
			self.clusters = []
		else:
			self.no_of_clusters_counter = 0
		#############################
		# Options about how to write clusters to disk.
		#############################
		# preparing folders to write information to for Collection 
		if path == None:
			self.path = os.getcwd()+'/'+self.name
		else:
			self.path = path
		self.have_database = have_database
		self.write_collection_history = write_collection_history
		if self.have_database or self.write_collection_history:
			self.make_collection_folder()
		#############################
		# set up database or folders for recording clusters to disk.
		# and if the database exists, make sure that it has been turned to read and write
		if self.have_database:
			self.database_path = self.path+'/'+str(self.name)+'.db'
		#############################
		# set up history writing folders for Collection
		if self.write_collection_history:
			self.create_collection_history()
		####################################################################################################################

	def __repr__(self):
		return '<Collection: '+str(self.name)+'> current size: '+str(len(self))+'; Clusters: '+str(self.clusters)

	def information(self):
		tostring =  '--------------------------------------------------------------------------\n'
		tostring += '<Collection: '+str(self.name)+'> current size: '+str(len(self))+'\n'
		for cluster in self.clusters:
			tostring += 'Cluster: '+str(cluster.name)+'; Energy: '+str(cluster.energy)+'; Fitness: '+str(cluster.fitness)+'\n'
		tostring += '--------------------------------------------------------------------------\n'
		return tostring

	def add_metadata(self):
		"""
		Due to an issue with some versions of ASE, this method will write metadata to the ASE database, if a database is being used to store cluster information.
		"""
		if self.have_database:
			metadata = {}
			metadata['title'] = str(self.name)
			metadata['default_columns'] = ['name', 'cluster_energy', 'id']
			# ----------------------------------------------------
			metadata['key_descriptions'] = {}
			metadata['key_descriptions']['name'] = ('Name', 'Name of the cluster.', '')
			metadata['key_descriptions']['gen_made'] = ('Generation Created', 'The generation the cluster was created.', '')
			metadata['key_descriptions']['cluster_energy'] = ('Cluster Energy', 'Potential energy of the cluster.', 'energy_units')
			metadata['key_descriptions']['ever_in_population'] = ('ever_in_population','This variable indicates if the cluster was ever in the population. If an offspring was made and was not ever accepted into the population, this variable will be False. If the cluster had been in the population for even one generation, this variable will be True.','bool')
			metadata['key_descriptions']['excluded_because_violates_predation_operator'] = ('excluded_because_violates_predation_operator','This variable will determine if a cluster was excluded from the population because it violated the predation operator.','bool')
			metadata['key_descriptions']['initial_population'] = ('initial_population','This variable will indicate if the cluster was apart of a newly created population.','bool')
			metadata['key_descriptions']['removed_by_memory_operator'] = ('removed_by_memory_operator','This variable will indicate if a cluster is removed because it resembles a cluster in the memory operator.','bool')
			# ----------------------------------------------------
			with connect(self.database_path) as database:
				database.metadata =  metadata.copy()
				database._metadata = metadata.copy()

	def make_collection_folder(self):
		"""
		Will create the directory for self.path if it does not exist.
		"""
		if not os.path.exists(self.path):
			os.mkdir(self.path)
		
	def __len__(self):
		"""
		Will return the number of clusters in the collection
		"""
		return len(self.clusters)

	def __getitem__(self,ith):
		"""
		Will return the cluster in the collection at the ith position.

		Inputs:
			ith (int): the index for the cluster that you want to get from the Collectino

		Returns:
			The cluster that you want to obtain from the Collections
		"""
		return self.clusters[ith]

	def __setitem__(self,ith,cluster):
		"""
		Will set a cluster at the ith position in the Collections.

		Inputs:
			ith (int): The index for the cluster that you want to get from the Collection
			cluster (Organisms.GA.Cluster): The Cluster that you want to set in the ith position of the Collection.

		"""
		self.clusters[ith] = cluster

	def __delitem__(self,ith):
		"""
		Remove the cluster at the ith position

		Inputs:
			ith (int): The index for the cluster that you want to get from the Collection

		"""
		self.remove(ith)

	# ---------------------------------------------------------------------------------------------------------------------------------------
	# methods to get names and energies of clusters

	def get_clusters(self):
		"""
		Returns all the clusters that the collection contains in a list

		Returns:
			all the clusters that the collection (list)
		"""
		return self.clusters

	def get_cluster_energies(self):
		"""
		Returns all the clusters that the collection contains in a list

		Returns:
			all the clusters that the collection (list)
		"""
		return [cluster.energy for cluster in self.clusters]

	def get_cluster_from_name(self,name):
		"""
		This method will return the cluster in the Collection with the name "name"

		Inputs:
			name (int): The name of the cluster you want to obtain from the Collection.

		Returns:
			The cluster in the Collection with the name "name" (Organisms.GA.Cluster)
		"""
		index = self.get_index(name)
		return self[index]

	def get_cluster_names(self,order=False):
		"""
		Will provide a list of all the names of all the clusters in the Collection

		Inputs:
			order (bool.): This tag will tell this method whether the user would like the list of names given in order. 

		Returns:
			List of the names of all the clusters in the Population
		"""
		if self.write_cluster_in_RAM:
			cluster_names = [cluster.name for cluster in self.clusters]
		else:
			import pdb; pdb.set_trace()
			exit('Cant do this')
		if order == True:
			cluster_names.sort()
		return cluster_names

	def sort_by_name(self):
		"""
		This method will sort the clusters in the list by their name.
		"""
		if self.write_cluster_in_RAM:
			self.clusters.sort(key=lambda cluster: cluster.name)
		else:
			import pdb; pdb.set_trace()
			exit('Cant do this')

	def sort_by_energy(self):
		"""
		This method will sort the clusters in the list by their energy (from lowest energy to highest energy).  
		"""
		if self.write_cluster_in_RAM:
			self.clusters.sort(key=lambda cluster: cluster.energy)
		else:
			import pdb; pdb.set_trace()
			exit('Cant do this')

	def get_index(self,name_to_find):
		"""
		This method will provide the index of the cluster that has the name "name_to_find" in the Collection

		Inputs:
			name_to_find (int): the name of the cluster in the Collection to obtain the index for

		Returns:
			the index of the cluster in the Collection with the name "name_to_find"

		Exceptions:
			Will break if the cluster with the name "name_to_find" can not be found in this method.
		"""
		if self.write_cluster_in_RAM:
			for index in range(len(self.clusters)):
				if self.clusters[index].name == name_to_find:
					return index
		else:
			print('Error in def get_index in class Collection, in Collection.py')
			print('You can use this method as the collection is not recording clusters in the RAM')
			print('Check your setting for write_cluster_in_RAM for '+str(self))
			exit('This program will exit without completing')
		print('Error in def get_index in Collection.py: Could not find the cluster that has the name = '+str(name_to_find))
		print('Check this')
		import pdb; pdb.set_trace()
		exit('This program will exit without completing.')

	def view_cluster(self,ith):
		"""
		Allow the user to visually look at the cluster using the ASE gui. This is a debugging method.

		Inputs:
			ith (int): the index of the ith cluster in the Collection to view in the ASE gui.
		"""
		self[ith].view()

	# ---------------------------------------------------------------------------------------------------------------------------------------
	# Add, remove, and replacing functions

	def add(self, index, cluster):
		"""
		Adds a cluster to the Collection.

		Index:
			index (int/str.): the index of the ith cluster in the Collection. If "End" is inputed, the cluster will be append to the end of the Collection list.
			cluster (Organisms.GA.Cluster): The cluster to add at the ith position in the Collection. 
		"""
		# Add the cluster to the Collection
		if self.write_cluster_in_RAM:	
			if index == 'End':
				self.clusters.append(cluster)
			else:
				self.clusters.insert(index,cluster)
		else:
			self.no_of_clusters_counter += 1
		# adds the cluster in the collection to disk
		if self.have_database:
			self.add_to_database(cluster)

	def remove(self, index):
		"""
		Removes a cluster to the Collection.

		Index:
			index (int): the index of the ith cluster in the Collection 
		"""
		# remove the cluster in the collection from the disk
		if self.have_database:
			self.remove_to_database(self.clusters[index])
		# Deletes the cluster from the Collection
		if self.write_cluster_in_RAM:
			del self.clusters[index]
		else:
			self.no_of_clusters_counter -= 1

	def replace(self, index, new_cluster):
		"""
		Will replace the ith cluster in the Collection with a new cluster. Uses the self.remove and self.add methods.

		Inputs:
			index (int): the index of the ith cluster in the Collection 
			new_cluster (Organisms.GA.Cluster): The new cluster to add at the ith position in the Collection
		"""
		self.remove(index)
		self.add(index, new_cluster)

	def pop(self,ith):
		"""
		Pops the cluster at the ith positiion of the Collection and returns it.

		Inputs:
			ith (int): The index for the cluster that you want to get from the Collection

		Returns:
			The cluster that you want to obtain from the Collections (Organisms.GA.Cluster)
		"""
		cluster = self.clusters[ith]
		del self[ith]
		return cluster

	# ---------------------------------------------------------------------------------------------------------------------------------------
	# methods to get the mean, max and min energies of clusters in the collection. 

	def mean_energy(self):
		"""
		The mean energy of the clusters in the population. 

		returns mean_energy: This is the mean energy of the cluster
		rtype   mean_energy: float
		"""
		cluster_energies = [cluster.energy for cluster in self.clusters]
		return mean(cluster_energies)

	def max_energy(self):
		"""
		The maximum energy of the clusters in the population. 

		returns maximum_energy: This is the maximum energy of the cluster
		rtype   maximum_energy: float
		"""
		cluster_energies = [cluster.energy for cluster in self.clusters]
		return max(cluster_energies)

	def min_energy(self):
		"""
		The minimum energy of the clusters in the population. 

		returns minimum_energy: This is the minimum energy of the cluster
		rtype   minimum_energy: float
		"""
		cluster_energies = [cluster.energy for cluster in self.clusters]
		return min(cluster_energies)

	def get_max_mean_min_energies(self):
		"""
		The maximum, mean, and minimum energy of the clusters in the population. 

		returns maximum_energy: This is the maximum energy of the cluster
		rtype   maximum_energy: float
		returns mean_energy: This is the mean energy of the cluster
		rtype   mean_energy: float
		returns minimum_energy: This is the minimum energy of the cluster
		rtype   minimum_energy: float
		"""
		cluster_energies = [cluster.energy for cluster in self.clusters]
		return max(cluster_energies), mean(cluster_energies), min(cluster_energies)

	def is_there_an_energy_range(self,rounding):
		"""
		Determines if there is a range of energies in the collection

		:param rounding: The rounding of the energy of the cluster
		:type  rounding: float

		returns Is there a range of energies in the collection
		rtype   bool
		"""
		cluster_energies = [cluster.energy for cluster in self.clusters]
		if round(max(cluster_energies),rounding) == round(min(cluster_energies),rounding):
			return False
		else:
			return True

	# ---------------------------------------------------------------------------------------------------------------------------------------
	# Methods for entering clusters into the database

	def add_to_database(self,cluster,center=False):
		"""
		Allows the user to write a cluster in the collection to a ASE database

		Inputs:
			cluster (Organisms.GA.Cluster): The cluster to add to the database.
		"""
		with connect(self.database_path) as database:
			cluster_to_database = cluster.deepcopy()
			if center:
				#unit_cell = cluster_to_database.get_cell()
				#cell_lengths = unit_cell.lengths()
				#middle_of_cell = cell_lengths/2.0
				cluster_to_database.center() #about=middle_of_cell)
			database.write(cluster_to_database,name=cluster.name,gen_made=cluster.gen_made,cluster_energy=cluster.energy,ever_in_population=cluster.ever_in_population,excluded_because_violates_predation_operator=cluster.excluded_because_violates_predation_operator,initial_population=cluster.initial_population,removed_by_memory_operator=cluster.removed_by_memory_operator)

	def remove_to_database(self,cluster):
		"""
		Allows the user to remove a cluster in the collection from the  ASE database

		Inputs:
			cluster (Organisms.GA.Cluster): The cluster to remove from the database.
		"""
		with connect(self.database_path) as database:
			cluster_id_to_remove = database.get(name=cluster.name).id
			del database[cluster_id_to_remove]

	def does_contain_database(self,backup):
		"""
		Does the collection contain a backup file.

		:param backup: If true, look for the backup database file. If false, look for the original database file. 
		:type  backup: bool.

		returns does_database_exist: Returns if either the database file or the backup database file exists. 
		rtype   does_database_exist: bool.

		"""
		if backup:
			return os.path.exists(self.database_path+'.backup.db')
		else:
			return os.path.exists(self.database_path)

	def delete_collection_database(self):
		"""
		This method will remove the whole database for this Collection from the disk.
		"""
		if os.path.exists(self.database_path):
			os.remove(self.database_path)

	# ----------------------------------------------------->
	# The following methods are for backing up the database.
	# ----------------------------------------------------->

	def backup_database(self):
		"""
		This method will make a backup of all the clusters in the Collection as a ASE database
		"""
		if self.have_database:
			copyfile(self.database_path, self.database_path+'.backup.db')

	def remove_backup_database(self):
		"""
		This method will remove the backup of all the clusters in the Collection, which will be in the format of a ASE database.
		"""
		if self.have_database:
			os.remove(self.database_path+'.backup.db')

	def remove_backup_database_if_exists(self):
		"""
		This method will remove the backup of all the clusters in the Collection, which will be in the format of a ASE database.
		"""
		if self.have_database:
			if os.path.exists(self.database_path+'.backup.db'):
				os.remove(self.database_path+'.backup.db')

	def move_backup_database_to_normal_backup(self):
		"""
		This method will remove the original database file and replace it with the backup database file. 
		"""
		if not self.have_database:
			return
		if os.path.exists(self.database_path):
			os.remove(self.database_path)
		copyfile(self.database_path+'.backup.db', self.database_path)
		os.remove(self.database_path+'.backup.db')

	# --------------------------------------------------------------------------------->
	# The following methods are for importing clusters from the database into the memory
	# --------------------------------------------------------------------------------->

	def import_clusters_from_database_to_memory(self, current_generation, clusters_in_resumed_population, clusters_in_resumed_population_energies, decimal_place):
		"""
		This method will attempt at obtaining the clustrs from the database and placing them in the collection in the RAM.

		This method is currently set up for reading the population, but if needed it can be reworked for general purpose.

		:param current_generation: The current generation
		:type  current_generation: float
		:param clusters_in_resumed_population: The names of the clusters in the current population 
		:type  clusters_in_resumed_population: list of int
		:param clusters_in_resumed_population_energies: The energies of the clusters in the current population 
		:type  clusters_in_resumed_population_energies: list of floats
		:param decimal_place: The number of decimal places that energies are rounded to in your genetic algorithm run
		:type  decimal_place: int

		returns did_clusters_come_from_backup: Did the imported clusters come from the backup database or the original. True if from the backup database, False if from the original backup.
		rtype   did_clusters_come_from_backup: bool.

		"""
		# determine if the collection wants to have clusters in its memory.
		use_backup_file = self.check_database_and_determine_if_to_use_backup()
		if not self.write_cluster_in_RAM:
			return 
		# method for loading clusters from a database and placing it into the RAM.
		def load_clusters_from_database_to_RAM(database_path,current_generation,clusters_in_resumed_population,clusters_in_resumed_population_energies,decimal_place,self):
			clusters = self.read_collection_database(database_path,current_generation)
			cluster_dict = {cluster.name: cluster for cluster in clusters}
			if self.check_clusters_in_database(cluster_dict, clusters_in_resumed_population, clusters_in_resumed_population_energies, decimal_place):
				self.add_clusters_into_RAM(cluster_dict, clusters_in_resumed_population)
			else:
				print('Error')
				import pdb; pdb.set_trace()
		# If the backup database exists, will gather all clusters from the backup, since this likely means that the last generation did not complete. 
		# Otherwise, will get clusters from the current database.
		if use_backup_file: # Getting data from current Current_Population_details.txt.backup and Population.db.backup.db
			path_to_database_db_to_use = self.database_path+'.backup.db'
		else:               # use data from the current Current_Population_details.txt and Population.db
			path_to_database_db_to_use = self.database_path
		# Load data from the collection database into the RAM
		load_clusters_from_database_to_RAM(self.database_path,current_generation,clusters_in_resumed_population,clusters_in_resumed_population_energies,decimal_place,self)
		# if use_backup_file = True,  data was obtained from backup files
		# if use_backup_file = False, data was obtained from original files
		return use_backup_file

	def check_database_and_determine_if_to_use_backup(self):
		"""
		This method will remove any journal or lock files associated with the database, as well as check that either the database or the backup is functional and can be used to allowing your genetic algorithm trial to resume. 
		
		returns did_use_backup_database: True means the backup database was used. False means the original database was used.
		rtype   did_use_backup_database: bool.
		"""
		database_files_to_check = (self.database_path+'-journal', self.database_path+'.lock', self.database_path+'.backup.db-journal', self.database_path+'.backup.db.lock')
		for database_file in database_files_to_check:
			if os.path.exists(database_file):
				os.remove(database_file)
		if (not os.path.exists(self.database_path+'.backup.db') and not os.path.exists(self.database_path)):
			print('Error in def check_database_and_determine_if_to_use_backup, class Collection, in Collection.py')
			error_message  = 'Population database error\n'
			error_message += 'Could not find neither\n'
			error_message += '--> '+str(self.database_path+'.backup.db')+'\n'
			error_message += '--> '+str(self.database_path)+'\n'
			error_message += 'Can not resume the genetic algorithm if the GA does not have either of these databases.'
			error_message += 'This algortihm will finish without completing.'
			exit(error_message)
		if os.path.exists(self.database_path+'.backup.db'):
			working, results = self.assess_clusters_in_database(self.database_path+'.backup.db')
			if working:
				print('The backup database is all good. Will use the backup database')
				return True # return True to say that the genetic algorithm will use the data from the backup population database
			print('There was an issue with clusters in the backup of the database.')
			print(str(results))
			print('Checking the original database')
		# something went wrong with the backup database file. Will attempt to get the clusters in the population from the original database. 
		working, results = self.assess_clusters_in_database(self.database_path)
		if working:
			print('The original database is all good. Will use the original database')
			return False # return False to say that the genetic algorithm will use the data from the original population database
		print('Error in def check_database_and_determine_if_to_use_backup, class Collection, in Collection.py')
		print('There was an issue with the last cluster:')
		print('----------------------------------------------')
		print(results)
		print('----------------------------------------------')
		print('Check this')
		exit('This program will exit without finishing.')

	def assess_clusters_in_database(self,database_path):
		"""
		This algorithm will check to make sure that there are no technical issues with the database, whether that is the original or the backup database.

		:param database_path: The path to the database
		:type  database_path: str.

		returns is_database_working: Is true if there are no technical issues with the database, False if there are technical issues.
		rtype   is_database_working: bool.
		returns reason_for_issue: Return a None object if everything is all good, otherwise returns the exception detailing the issues with the database or the name of the cluster that caused issues. 
		rtype   is_database_working: None or str.

		"""
		cluster_names_in_database = []
		try:
			with connect(database_path) as database:
				for cluster_row in database.select():
					ase_cluster = cluster_row.toatoms()
					ase_cluster_name = cluster_row.name
					if ase_cluster_name in cluster_names_in_database:
						return (False, ase_cluster_name)
					cluster_names_in_database.append(ase_cluster_name)
		except Exception as exception:
			return (False, exception)
		if not len(cluster_names_in_database) == self.size:
			return (False, Exception('Not enough clusters in the database.\nDatabase size: '+str(len(cluster_names_in_database))+'\nSize of cluster: '+str(self.size)+'.'))
		return (True, None)

	def read_collection_database(self,database_path,current_generation=None):
		"""
		This method will read the clusters in the database. Furthermore, this method is also designed to repair the collection database by removing any clusters that were created after the current generation if desired.

		:param database_path: The path to the database. 
		:type  database_path: str. 
		:param current_generation: The current generation that your genetic algorithm trial is being resumed from
		:type  current_generation: int

		returns clusters: This is a list of all the clusters from the database
		rtype   clusters: list of Organisms.GA.Clusters

		"""
		# Get clusters from database
		clusters = []
		cluster_to_delete = []
		with connect(database_path) as database:
			for cluster_row in database.select():
				if current_generation == None or cluster_row.gen_made <= current_generation:
					ase_cluster = cluster_row.toatoms()
					cluster = Cluster(ase_cluster)
					cluster.custom_verify_cluster(cluster_row.name,cluster_row.gen_made,cluster_row.cluster_energy,cluster_row.ever_in_population,cluster_row.excluded_because_violates_predation_operator,cluster_row.initial_population)
					clusters.append(cluster)
				else:
					cluster_to_delete.append(cluster.id)
		database.delete(cluster_to_delete)
		return clusters

	def remove_clusters_from_database_that_are_from_unsuccessful_generations(self,database_path,current_generation=None):
		"""
		This method will go through the database and delete any clusters of unsuccessful generations. 

		:param database_path: The path to the database. 
		:type  database_path: str. 
		:param current_generation: The current generation that your genetic algorithm trial is being resumed from
		:type  current_generation: int

		"""
		if current_generation == None:
			return
		with connect(database_path) as database:
			cluster_to_delete = []
			for cluster_row in database.select():
				if cluster_row.gen_made > current_generation:
					cluster_to_delete.append(cluster_row.gen_made)
			database.delete(cluster_to_delete)
		
	def check_clusters_in_database(self, cluster_dict, cluster_names, cluster_energies, decimal_place):
		"""
		This method will check that the database contains all the clusters you need and does not have any issues.

		:param cluster_dict: This is a dicionary of all the clusters from the database, given as {cluster_name: Cluster}
		:type  cluster_dict: {int: ASE.Cluster}
		:param cluster_names: list of the names of the clusters that are needed for the collection
		:type  cluster_names: list of int
		:param cluster_energies: list of the energies of the clusters that are needed for the collection
		:type  cluster_energies: list of float

		returns is_database_all_good: True means the database is all good and contains all the clusters needed to restore the collection, as well as comfirming they are of the correct energy. False means something is not working with the database, the database does not contain a required cluster, or there is an issue with clusters not having the energy that they should have. 
		rtype   is_database_all_good: bool.
		"""
		try:
			for cluster_name, cluster_energy in zip(cluster_names,cluster_energies):
				if not cluster_name in cluster_dict.keys():
					print('Note for clusters in cluster: Can not find cluster '+str(cluster_name)+' in the Population database')
					print('Cluster that can not be found: '+str(cluster_name))
					print('Clusters in the population database: '+str(cluster_dict.keys()))
					return False
				cluster = cluster_dict[cluster_name]
				if not round(cluster.energy,decimal_place) == round(cluster_energy,decimal_place):
					print('Found a cluster in the population that has a different energy to that cluster as recorded in the database.')
					print('Energy of cluster as recorded in EnergyProfile.txt: '+str(round(cluster_energy,decimal_place)))
					print('Energy of cluster as recorded in the database: '+str(round(cluster.energy,decimal_place)))
					return False
			return True
		except:
			return False

	def add_clusters_into_RAM(self, cluster_dict, cluster_names):
		"""
		This method adds clusters into the RAM

		:param cluster_dict: This is a dicionary of all the clusters from the database, given as {cluster_name: Cluster}
		:type  cluster_dict: {int: ASE.Cluster}
		:param cluster_names: list of the names of the clusters that are needed for the collection
		:type  cluster_names: list of int
		"""
		if self.write_cluster_in_RAM:	
			for cluster_name in cluster_names:
				cluster = cluster_dict[cluster_name]
				self.clusters.append(cluster)

	# ---------------------------------------------------------------------------------------------------------------------------------------
	############################# HISTORY FILE METHODS #############################
	# The history file is mainly created for the Population to keep track of the   #
	# clusters that were made during the Organisms program.                           #
	################################################################################

	def history_file_name(self,end_name=None):
		"""
		Get the name of the history file for this Collection.

		Inputs:
			end_name (str): The suffix of the name for the history gfile.
		"""
		if end_name == None:
			self.history_path = self.path+'/'+str(self.name)+"_history.txt"
		else:
			self.history_path = self.path+'/'+str(self.name)+"_history_"+str(end_name)+".txt"

	def get_history_path(self):
		"""
		Return the path to the history file

		Returns:
			the path to the history file
		"""
		return self.history_path

	def open(self,w_or_a):
		"""
		This opens the profile pool text file

		Inputs:
			w_or_a = 'Indicates how to open the file, whether to open it as a new file ("w") or to append information to the history file ("a").
		"""
		self.history_file = open(self.get_history_path(),str(w_or_a))

	def close(self):
		"""
		This closes the history text file.
		"""
		self.history_file.close()
		del self.history_file

	def create_collection_history(self):
		"""
		This definition will create the history file. 

		This includes the folder, contents of the folder and beginning to write the history file.

		It also included information about the clusters in the collection file when it was first created.
		"""
		if not self.write_collection_history:
			return
		self.history_file_name()
		if os.path.exists(self.get_history_path()):
			return
		self.open('w')
		self.history_file.write("This file records all the information about the clusters in the pool and other details about the clusters in the pool. \n")
		self.history_file.write("--------------------------------\n")
		self.history_file.write("OVERALL DETAILS\n")
		self.history_file.write("Collection name: "+str(self.name)+"\n")
		#self.profile.write("Cluster: ")
		#for element in self.cluster_makeup:
		#	self.profile.write(str(element) + " " + str(self.cluster_makeup[element]) + " ")
		#self.profile.write("\n")
		self.history_file.write("--------------------------------\n")
		self.close()

	def add_to_history_file(self,generation_number,is_epoch=False,epoch_due_to_population_energy_convergence=None):
		"""
		This definition will add the information of the population. This is suppose to be used after each generation has completed. 

		:param generation_number: The current generation that the genetic algorithm run has just performed.
		:type  generation_number: int 
		:param is_epoch: Has an epoch just occurred
		:type  is_epoch: bool.
		:param epoch_due_to_population_energy_convergence: If an epoch occurred, was it because the energy of the clusters converged.
		:type  epoch_due_to_population_energy_convergence: bool.

		"""
		if not self.write_collection_history:
			return
		self.open('a')
		if is_epoch:
			self.history_file.write('#-----------------------------------------------------------------------------\n')
			if epoch_due_to_population_energy_convergence:
				self.history_file.write('# -> Reset populationas the clusters in the population converged upon the same energy\n')
			else:
				self.history_file.write('# -> Reset population as reached epoch\n')
			self.history_file.write('#-----------------------------------------------------------------------------\n')
		self.history_file.write("GA Iteration: " + str(generation_number) + "\n")
		self.history_file.write("Clusters in Pool:\t")
		for cluster in self:
			self.history_file.write(str(cluster.name)+"("+str(cluster.gen_made)+")\t")
		self.history_file.write("\n")
		self.history_file.write("Energies of Clusters:\t")
		for cluster in self:
			self.history_file.write(str(cluster.energy) + "\t")
		self.history_file.write("\n")
		self.close()

	def check_PoolProfileTXT_exists(self):
		"""
		This definition checks to see if the PoolProfile folder exists
		"""
		return os.path.exists(self.history_file) and os.path.isfile(self.history_file)

	def check_historyfile(self,resume_from_generation):
		'''
		This method will check the history file to make sure that it does not contain any information from a failed generation.

		If it does contain information from failed generations, it will delete those lines from the history file.

		:param generation_number: The current generation that the genetic algorithm run has just performed.
		:type  generation_number: int 

		'''
		while True:
			counter = 0
			lines_to_remove = 0
			last_lines_in_historyfile = tail(self.history_path,9)[::-1]
			for line in last_lines_in_historyfile:
				line = line.decode("utf-8") 
				if line.startswith('GA Iteration:'):
					line = line.rstrip()
					generation = int(line.replace('GA Iteration:',''))
					if generation > resume_from_generation:
						counter += 1
						lines_to_remove = counter
					elif generation == resume_from_generation:
						break
					elif generation < resume_from_generation:
						print('Error in def check_historyfile, in class Collection, in Collection.py')
						print('The last generation in the '+str(self.history_path)+' is less than the resumed generation.')
						print('Check the Population_history file for this GA run before continuing. There may be an issue here')
						print('Last generation in the Population_history file: '+str(generation))
						print('Generation to resume from: '+str(resume_from_generation))
						import pdb; pdb.set_trace()
						exit('The genetic algorithm will finished without starting.')
				else:
					counter += 1
			if lines_to_remove > 0:
				print('Removing the last '+str(lines_to_remove)+' from '+str(self.history_path))
				remove_end_lines_from_text(self.history_path, lines_to_remove)
			else:
				break
		last_lines_in_historyfile = tail(self.history_path,3)
		historyfile_generation = int(last_lines_in_historyfile[0].decode("utf-8").rstrip().replace('GA Iteration:',''))
		if historyfile_generation == resume_from_generation:
			return
		else:
			print('Error in def check_historyfile, in class Collection, in Collection.py')
			print('The last generation in the '+str(self.history_path)+' is less than the resumed generation.')
			print('Check the Population_history file for this GA run before continuing. There may be an issue here')
			print('Last generation in the Population_history file: '+str(historyfile_generation))
			print('Generation to resume from: '+str(resume_from_generation))
			import pdb; pdb.set_trace()
			exit('The genetic algorithm will finished without starting.')
